#include <stdlib.h>
#include <cmath>
#include <iostream>
#include <fstream>
using namespace std;

const double GRAD2DEG = 180.0/M_PI;
inline double rad(double grad) { return grad/GRAD2DEG; }
inline double grad(double rad)  { return  rad*GRAD2DEG; }

void usage(char *pgrname)
{
  cout << "USAGE:" << endl;
  cout << "  " << pgrname << " [<step>]   (take only every step-th scan for map generation)" << endl;
  cout << "  e.g. " << pgrname << "10" << endl;
  exit(1);
}


int main(int argc,char **argv) {

  ifstream in, in_odom;
  ofstream out, gnu("map.gnu");
  char infile[400];
  char outfile[400];
  const char* in_odomfile = "pose@scan.dat";

  int nr = 0;
  char dummy[400];
  double angle, x, z;              // scan data
  int dist;                        // scan data
  double x_transf, z_transf;       // transformed scan data
  double x_coord, z_coord, theta;  // from odometry

  int step = 1;                    // process only every step-th file!
  if (argc == 2) step = atoi(argv[1]);
  if (step < 1 || argc > 2) usage(argv[0]);

  in_odom.open(in_odomfile);
  if (!in_odom.good()) {
    cerr << "Error reading input file \"" << in_odomfile << "\"." << endl;
    exit(1);
  }

  // generate gnuplot file for map output
  gnu << "set grid" << endl
	 << "set pointsize .3" << endl
	 << "set xrange[-1400:1400]" << endl
	 << "set yrange[-800:2000]" << endl
	 << "set xlabel \"[cm]\"" << endl
	 << "set ylabel \"[cm]\"" << endl
	 << "set title \"Odometry based map\"" << endl
	 << "plot ";
  
  // could've been done more efficiently, but this while construct is just too much fun... ;-)
  while (sprintf(infile, "scans/Front%0.6d.dat", nr++), in.open(infile), in.good()) {

    // read odometry data
    if (!in_odom.good()) { cerr << "Error: Too few odometry data lines!\n"; exit(2); }
    in_odom >> x_coord >> z_coord >> theta;
    
    if ((nr-1) % step  != 0) { in.close(); continue; }

    sprintf(outfile, "scans/transformed/Front%0.6d.dat", nr-1);
    out.open(outfile);
    if (!out.good()) {
	 cerr << "ERROR writing to " << outfile << endl;
	 exit(1);
    }

    cout << "Transform file " << infile << " by ("
	    << x_coord << " : " << z_coord << " : " << grad(theta) << ")\n";

    // read scan data, transform & write
    in.getline(dummy, 400);  // read first 2 comment lines
    in.getline(dummy, 400);
    while (in.good()) { 
	 in >> angle >> dist >> x >> z;

	 // transform
	 x_transf =   x * cos(theta) + z * sin(theta) + x_coord;
	 z_transf = - x * sin(theta) + z * cos(theta) + z_coord;

	 // store
	 out << x_transf << " " << z_transf << endl;  // don't need angle & dist
    }

    // generating map (gnuplot)
    if (nr < 2)
	 gnu << "\"" << outfile << "\" t \"Scan data (every " << step << "-th scan)" << "\" w p 1 ";
    else
	 gnu <<  ",\\\n     \"" << outfile << "\" t \"\" w p 1 ";

    in.close(); in.clear();
    out.close(); out.clear();
  }

  gnu << ",\\\n     \"" << in_odomfile << "\" t \"Trajectory\" w p 3 " << endl;
  gnu << "pause -1" << endl;

  system("gnuplot map.gnu");
  return 0;
}
